/*  Autopackage Superuser
 *  Copyright (C) 2004  Hongli Lai <h.lai@chello.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gtk/gtk.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include "autosu.h"
#include "su-icon.c"


/********************************************
 * Utility functions
 ********************************************/

static void
clear_entry (GtkWidget *entry)
{
	gchar *blank;

	/* Make a pathetic stab at clearing the GtkEntry field memory */
	blank = (gchar *) gtk_entry_get_text (GTK_ENTRY (entry));
	if (blank && strlen (blank))
		memset (blank, ' ', strlen (blank));

	blank = g_strdup (blank);
	if (strlen (blank))
		memset (blank, ' ', strlen (blank));

	gtk_entry_set_text (GTK_ENTRY (entry), blank);
	gtk_entry_set_text (GTK_ENTRY (entry), "");
}


void
show_error (gchar *format, ...)
{
	GtkWidget *dialog;
	va_list ap;
	gchar *msg;

	va_start (ap, format);
	msg = g_strdup_vprintf (format, ap);
	va_end (ap);

	gdk_threads_enter ();
	dialog = gtk_message_dialog_new (NULL,
		GTK_DIALOG_MODAL,
		GTK_MESSAGE_ERROR,
		GTK_BUTTONS_CLOSE,
		"%s", msg);
	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
	gdk_threads_leave ();
}


/********************************************
 * GUI code
 ********************************************/

struct _GUI {
	guint pass;
	gchar *user;
	gchar *command;
	GtkWidget *win;
	GtkWidget *entry;
	GtkWidget *status;
	GtkWidget *nopasswd;
	GtkWidget *cancel;
	GtkWidget *ok;
} GUI;


static void
cb_no_password ()
{
	exit (EXIT_NOPASSWD);
}


static void
cb_cancel ()
{
	exit (EXIT_CANCEL);
}


static gpointer start_check (gpointer data)
{
	gchar *password;
	gboolean result;

	password = data;
	result = su_check_password (GUI.user, password);

	if (result)
	{
		/* Correct password; run command and return */
		char *ret;

		signal (SIGINT, SIG_DFL);
		gdk_threads_enter ();
		gtk_widget_destroy (GUI.win);
		gdk_threads_leave ();

		ret = g_strdup_printf ("%d", su_exec (GUI.user, GUI.command, password));
		write (5, ret, strlen (ret));
		exit (0);
	} else {
		/* Incorrect password; exit if the user typed incorrect password 3 times */
		gchar *tmp;

		safefree (password);
		if (GUI.pass >= 3)
			exit (EXIT_INCORRECT);

		gdk_threads_enter ();
		if (GUI.pass == 2)
			tmp = g_strdup_printf ("<b>%s</b>",
				_("Password incorrect, please try again. You have one more chance."));
		else
			tmp = g_strdup_printf ("<b>%s</b>",
				_("Password incorrect, please try again."));
		gtk_label_set_markup (GTK_LABEL (GUI.status), tmp);
		g_free (tmp);

		gtk_widget_set_sensitive (GUI.entry, TRUE);
		if (GUI.nopasswd) gtk_widget_set_sensitive (GUI.nopasswd, TRUE);
		gtk_widget_set_sensitive (GUI.cancel, TRUE);
		gtk_widget_set_sensitive (GUI.ok, TRUE);
		gtk_widget_grab_focus (GUI.entry);
		gdk_threads_leave ();
	}
	return NULL;
}


static void
cb_ok ()
{
	gchar *tmp, *password;

	password = g_strdup (gtk_entry_get_text (GTK_ENTRY (GUI.entry)));
	clear_entry (GUI.entry);

	tmp = g_strdup_printf ("<b>%s</b>",
		"Checking password, please wait...");
	gtk_label_set_markup (GTK_LABEL (GUI.status), tmp);
	g_free (tmp);

	gtk_widget_show (GUI.status);
	gtk_widget_set_sensitive (GUI.entry, FALSE);
	if (GUI.nopasswd) gtk_widget_set_sensitive (GUI.nopasswd, FALSE);
	gtk_widget_set_sensitive (GUI.cancel, FALSE);
	gtk_widget_set_sensitive (GUI.ok, FALSE);

	GUI.pass++;
	g_thread_create (start_check, password, FALSE, NULL);
}


static gint
ask_password (gchar *user, gchar *command, Options *options)
{
	GtkWidget *win;
	GtkWidget *vbox,
		*hbox,
			*image,
			*intro,
		*table,
			/* *label1, *commandlabel, */
			*label2, *entry,
		*status,
		*hbox2,
			*btnbox1,
				*nopassword,
			*hbox3,
				*btnbox2,
					*cancel,
				*btnbox3,
					*ok;
	GtkWidget *btnalign, *btnhbox, *btnimage, *btnlabel;
	GtkTooltips *tips;
	GdkPixbuf *pixbuf;
	gboolean is_kde;
	gchar *tmp;
	gchar *title = NULL, *header = NULL, *description = NULL,
	      *ok_label = NULL, *cancel_label = NULL, *no_password_label = NULL;

	is_kde = !system ("ps ux | grep -v grep | grep -qE '(kicker|slicker|karamba)'");

	memset (&GUI, 0, sizeof (GUI));
	GUI.user = user;
	GUI.command = command;

	/* Mode-specific labels */
	switch (options->mode)
	{
	case APKG_INSTALL:
		title = _("autopackage support code installation");
		description = _("If yes, please enter it below. The autopackage support code will then be installed system-wide.");
		break;
	default:
		break;
	}
	
	/* Default labels */
	if (!title)
	{
		if (options->displayName)
			title = g_strdup_printf (_("%s - Software installation"), options->displayName);
		else
			title = _("Software installation");
	}
	if (!header) header = _("Do you have the superuser (root) password?");
	if (!description)
	{
		if (options->displayName)
			description = g_strdup_printf (
				_("If yes, please enter it below. \"%s\" will then be installed system-wide."),
				options->displayName);
		else
			description = _("If yes, please enter it below.");
	}
	if (!ok_label) ok_label = _("C_ontinue");
	if (!cancel_label) cancel_label = _("_Cancel");
	if (!no_password_label) no_password_label = _("_No Password");


	GUI.win = win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_container_set_border_width (GTK_CONTAINER (win), 12);
	gtk_window_set_title (GTK_WINDOW (win), title);
	gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER);
	gtk_window_set_resizable (GTK_WINDOW (win), FALSE);
	g_signal_connect (win, "delete_event", G_CALLBACK (gtk_true), NULL);

        // setup layout
	vbox = gtk_vbox_new (FALSE, 12);
	gtk_container_add (GTK_CONTAINER (win), vbox);
	hbox = gtk_hbox_new (FALSE, 12);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);

        // set window icon and main image
	pixbuf = gdk_pixbuf_new_from_inline (sizeof (su_icon),
			(const guint8 *) su_icon, FALSE, NULL);
	image = gtk_image_new_from_pixbuf (pixbuf);
	gtk_window_set_icon (GTK_WINDOW (win), pixbuf);
	g_object_unref (pixbuf);
	gtk_box_pack_start (GTK_BOX (hbox), image, FALSE, FALSE, 0);

        // intro text
	intro = gtk_label_new ("");
	gtk_label_set_line_wrap (GTK_LABEL (intro), TRUE);
	tmp = g_strdup_printf ("<b>%s</b>\n%s", header, description);
	gtk_label_set_markup (GTK_LABEL (intro), tmp);
	g_free (tmp);
	gtk_misc_set_alignment (GTK_MISC (intro), 0.0, 0.0);
	gtk_label_set_selectable (GTK_LABEL (intro), TRUE);
	gtk_box_pack_start (GTK_BOX (hbox), intro, TRUE, TRUE, 0);

        // more layout stuff
	table = gtk_table_new (2, 2, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE (table), 6);
	gtk_table_set_row_spacings (GTK_TABLE (table), 6);
	gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, FALSE, 0);

/*	label1 = gtk_label_new (_("Command to run:"));
	gtk_misc_set_alignment (GTK_MISC (label1), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), label1, 0, 1, 0, 1,
		GTK_FILL, 0,
		0, 0);

	commandlabel = gtk_label_new (command);
	gtk_misc_set_alignment (GTK_MISC (commandlabel), 0.0, 0.5);
	gtk_label_set_selectable (GTK_LABEL (commandlabel), TRUE);
	gtk_table_attach (GTK_TABLE (table), commandlabel, 1, 2, 0, 1,
		GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		0, 0); */

	label2 = gtk_label_new (_("Root password:"));
	gtk_misc_set_alignment (GTK_MISC (label2), 1.0, 0.5);
	gtk_table_attach (GTK_TABLE (table), label2, 0, 1, 1, 2,
		GTK_FILL, 0,
		0, 0);

        // password entry
	GUI.entry = entry = gtk_entry_new ();
	gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
	gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
	gtk_table_attach (GTK_TABLE (table), entry, 1, 2, 1, 2,
		GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL,
		0, 0);
	gtk_widget_grab_focus (entry);

        // status label
	GUI.status = status = gtk_label_new ("");
	tmp = g_strdup_printf ("<b>%s</b>",
		"Checking password, please wait...");
	gtk_label_set_markup (GTK_LABEL (status), tmp);
	gtk_label_set_selectable (GTK_LABEL (status), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox), status, FALSE, FALSE, 0);
	g_free (tmp);

	hbox2 = gtk_hbox_new (FALSE, 12);
	gtk_box_pack_start (GTK_BOX (vbox), hbox2, FALSE, FALSE, 0);

        // buttons
	btnbox1 = gtk_hbutton_box_new ();
	gtk_button_box_set_layout (GTK_BUTTON_BOX (btnbox1), GTK_BUTTONBOX_START);
	gtk_box_pack_start (GTK_BOX (hbox2), btnbox1, TRUE, TRUE, 0);

        // add no password button
        GUI.nopasswd = nopassword = NULL;
        if (!options->rootonly) {
                GUI.nopasswd = nopassword = gtk_button_new ();
                gtk_container_add (GTK_CONTAINER (btnbox1), nopassword);
                
                btnalign = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
                gtk_container_add (GTK_CONTAINER (nopassword), btnalign);
                
                btnhbox = gtk_hbox_new (FALSE, 2);
                gtk_container_add (GTK_CONTAINER (btnalign), btnhbox);
                btnimage = gtk_image_new_from_stock (GTK_STOCK_QUIT, GTK_ICON_SIZE_BUTTON);
                gtk_box_pack_start (GTK_BOX (btnhbox), btnimage, FALSE, FALSE, 0);
                btnlabel = gtk_label_new (no_password_label);
                gtk_label_set_use_underline (GTK_LABEL (btnlabel), TRUE);
                gtk_box_pack_start (GTK_BOX (btnhbox), btnlabel, FALSE, FALSE, 0);
                g_signal_connect (nopassword, "clicked", G_CALLBACK (cb_no_password), NULL);
        }
        
	hbox3 = gtk_hbox_new (FALSE, 6);
	gtk_box_pack_start (GTK_BOX (hbox2), hbox3, TRUE, TRUE, 0);

	btnbox2 = gtk_hbutton_box_new ();
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (btnbox2), 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (btnbox2), GTK_BUTTONBOX_END);

        // cancel button
	GUI.cancel = cancel = gtk_button_new ();
	btnalign = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
	gtk_container_add (GTK_CONTAINER (cancel), btnalign);
	btnhbox = gtk_hbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (btnalign), btnhbox);
	btnimage = gtk_image_new_from_stock (GTK_STOCK_CANCEL, GTK_ICON_SIZE_BUTTON);
	gtk_box_pack_start (GTK_BOX (btnhbox), btnimage, FALSE, FALSE, 0);
	btnlabel = gtk_label_new (cancel_label);
	gtk_label_set_use_underline (GTK_LABEL (btnlabel), TRUE);
	gtk_box_pack_start (GTK_BOX (btnhbox), btnlabel, FALSE, FALSE, 0);
	gtk_container_add (GTK_CONTAINER (btnbox2), cancel);
	g_signal_connect (cancel, "clicked", G_CALLBACK (cb_cancel), NULL);

        // ok button
	btnbox3 = gtk_hbutton_box_new ();
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (btnbox3), 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (btnbox3), GTK_BUTTONBOX_END);

	GUI.ok = ok = gtk_button_new ();
	btnalign = gtk_alignment_new (0.5, 0.5, 0.0, 0.0);
	gtk_container_add (GTK_CONTAINER (ok), btnalign);
	btnhbox = gtk_hbox_new (FALSE, 2);
	gtk_container_add (GTK_CONTAINER (btnalign), btnhbox);
	btnimage = gtk_image_new_from_stock (GTK_STOCK_OK, GTK_ICON_SIZE_BUTTON);
	gtk_box_pack_start (GTK_BOX (btnhbox), btnimage, FALSE, FALSE, 0);
	btnlabel = gtk_label_new (ok_label);
	gtk_label_set_use_underline (GTK_LABEL (btnlabel), TRUE);
	gtk_box_pack_start (GTK_BOX (btnhbox), btnlabel, FALSE, FALSE, 0);
	gtk_container_add (GTK_CONTAINER (btnbox3), ok);
	GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
	g_signal_connect (ok, "clicked", G_CALLBACK (cb_ok), NULL);

	/* Reverse button order if we're running in KDE */
	if (is_kde)
	{
		gtk_box_pack_start (GTK_BOX (hbox3), btnbox3, TRUE, TRUE, 0);
		gtk_box_pack_start (GTK_BOX (hbox3), btnbox2, FALSE, TRUE, 0);
	} else
	{
		gtk_box_pack_start (GTK_BOX (hbox3), btnbox2, TRUE, TRUE, 0);
		gtk_box_pack_start (GTK_BOX (hbox3), btnbox3, FALSE, TRUE, 0);
	}
	gtk_widget_grab_default (ok);


	tips = gtk_tooltips_new ();
        if (!options->rootonly)
                gtk_tooltips_set_tip (tips, nopassword,
                                      _("Click here if you do not know the root password, or if you want to install the software locally (as normal user)."), NULL);
	gtk_tooltips_set_tip (tips, cancel, _("Cancel the installation process."), NULL);
	gtk_tooltips_set_tip (tips, ok, _("Verify the password and continue installing the software system-wide."), NULL);


	gtk_widget_show_all (hbox);
	gtk_widget_show_all (table);
	gtk_widget_show_all (hbox2);
	gtk_widget_show (vbox);
	gtk_widget_show (win);
	gdk_threads_enter ();
	gtk_main ();
	gdk_threads_leave ();
	return 0;
}


static void
sigint_handler (int num)
{
	exit (EXIT_CANCEL);
}


int main (int argc, char *argv[])
{
	gchar *user = "root";
	gchar *command;
	Options options;

	memset (&options, 0, sizeof (Options));
	command = create_command (parse_args (argc, argv, &options));
	su_check_pam (&(argv[1]));

	signal (SIGINT, sigint_handler);

	g_thread_init (NULL);
	gdk_threads_init ();
	gtk_init (&argc, &argv);

	if (!su_needs_password (user))
	{
		char *ret;

		signal (SIGINT, SIG_DFL);
		ret = g_strdup_printf ("%d", su_exec (user, command, NULL));
		write (5, ret, strlen (ret));
		exit (0);
	}

	ask_password (user, command, &options);
        g_assert_not_reached();
        return 1;
}
