/*  Autopackage Superuser
 *  Copyright (C) 2004  Hongli Lai <h.lai@chello.nl>
 *                      Mike Hearn <mike@navi.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <termios.h>
#include <dlfcn.h>
#include "autosu.h"


static FILE *tty = NULL;
static int consoleWidth = 0;


void
show_error (gchar *format, ...)
{
	va_list ap;
	gchar *msg;

	va_start (ap, format);
	msg = g_strdup_vprintf (format, ap);
	va_end (ap);
	fputs ("\e[1;31m", tty);
	fprintf (tty, _("Error: %s\n"), msg);
	fputs ("\e[0m", tty);
	fflush (tty);
}


static void
printline (char *text)
{
	int len, i;

	fputc (' ', tty);
	fputs (text, tty);
	len = strlen (text) + 1;
	if (len < consoleWidth)
		for (i = 0; i < consoleWidth - len; i++)
			fputc (' ', tty);
	fputc ('\n', tty);
}


static void
printchoices (int invalid, Options *options)
{
	int i;

	setvbuf (tty, NULL, _IOFBF, 0);

	fputs ("\e[1;37m", tty);
	for (i = 0; i < consoleWidth; i++)
		fputc ('_', tty);
	fputc ('\n', tty);
	fputs ("\e[1;44m\e[1;37m", tty);
	printline ("");

	if (!invalid)
	{
		char *display;

		if (options->mode == APKG_INSTALL)
		{
			printline ("[ autopackage support code installation ]");
			printline ("");
		} else if (options->displayName)
		{
			display = g_strdup_printf (_("[ Installing: %s ]"), options->displayName);
			printline (display);
			g_free (display);
			printline ("");
		}

		fputs ("\e[1;33m", tty);
                if (options->rootonly)
                        printline(_("You need to know the superuser (root) password for this package."));
                else
                        printline (_("Do you know the superuser (root) password?"));
		fputs ("\e[1;37m", tty);

                if (options->rootonly) printline(_("Please enter it now."));
                else printline(_("If so, please enter it."));
		printline ("");
	}

	fputs ("\e[1;36m", tty);

	printline (_("1. Yes, continue with password (install system-wide)."));
        if (!options->rootonly) {
                printline (_("2. No, I don't know the root password (install as normal user)."));
                printline (_("3. Cancel installation."));
        } else printline (_("2. Cancel installation."));
               
	fputs ("\e[1;37m", tty);

	for (i = 0; i < consoleWidth; i++)
		fputc ('_', tty);
	fprintf (tty, "\033[0m\n");

	fputs (_("Enter the number of a choice and press ENTER: "), tty);

	fflush (tty);
	setlinebuf (tty);
}


static void
ask_password (char *user, char *command, Options *options)
{
	char line[64];
	int i;

	printchoices (0, options);
	fgets (line, sizeof (line), stdin);

	while (line[0] != '1' && line[0] != '2' && line[0] != '3')
	{
		fprintf (tty, "\n");
		if (line[0] == '\n')
			show_error (_("Please enter a choice."));
		else
			show_error (_("Invalid choice '%c'."), line[0]);
		printchoices (1, options);
		fgets (line, sizeof (line), stdin);
	}

	if ((line[0] == '2') && !options->rootonly)
		exit (EXIT_NOPASSWD);
        else if ((line[0] == '2') && options->rootonly)
                exit (EXIT_CANCEL);
	else if (line[0] == '3')
		exit (EXIT_CANCEL);

	for (i = 0; i < 3; i++)
	{
		char *result, *password;
		gboolean valid;
		struct termios ios;

		result = getpass (_("Root password: "));
		if (!result)
		{
			show_error (strerror (errno));
			exit (EXIT_SYSERROR);
		}

		password = g_strdup (result);
		memset (result, 0, _SC_PASS_MAX);
		valid = su_check_password (user, password);
		/* Restore terminal ECHO mode */
		tcgetattr (1, &ios);
		ios.c_lflag |= ECHO;
		tcsetattr (1, TCSANOW, &ios);

		if (valid)
		{
			char *ret;

			signal (SIGINT, SIG_DFL);
			ret = g_strdup_printf ("%d", su_exec (user, command, password));
			write (5, ret, strlen (ret));
			exit (0);
		} else if (i == 0)
			show_error (_("Incorrect password, please try again."));
		else if (i == 1)
			show_error (_("Incorrect password, please try again. You have one more chance."));
	}
	exit (EXIT_INCORRECT);
}


static void
sigint_handler (int num)
{
	exit (EXIT_CANCEL);
}


int main (int argc, char *argv[])
{
	gchar *user = "root";
	gchar *command;
	Options options;
	void *ncurses;

	memset (&options, 0, sizeof (Options));
	command = create_command (parse_args (argc, argv, &options));
	su_check_pam (&(argv[1]));

	tty = fopen ("/dev/tty", "w");
	if (!tty)
		return EXIT_SYSERROR;

	signal (SIGINT, sigint_handler);

	if (!su_needs_password (user))
	{
		char *ret;

		signal (SIGINT, SIG_DFL);
		ret = g_strdup_printf ("%d", su_exec (user, command, NULL));
		write (5, ret, strlen (ret));
		exit (0);
	}

	ncurses = dlopen ("libncurses.so.5", RTLD_LAZY);
	if (!ncurses)
		ncurses = dlopen ("libncurses.so.4", RTLD_LAZY);
	if (!ncurses)
		ncurses = dlopen ("libncurses.so", RTLD_LAZY);
	if (ncurses)
	{
		void (*initscr)() = dlsym (ncurses, "initscr");
		void (*endwin)() = dlsym (ncurses, "endwin");
		int *COLS = dlsym (ncurses, "COLS");
		if (initscr && endwin && COLS)
		{
			initscr ();
			consoleWidth = *COLS;
			endwin ();
		}
		dlclose (ncurses);
	}

	if (!consoleWidth && getenv ("COLUMNS"))
		consoleWidth = atoi (getenv ("COLUMNS"));
	if (!consoleWidth) consoleWidth = 80;

	ask_password (user, command, &options);
	return 0;  /* Never reached */
}
