/*  Autopackage Superuser
 *  Copyright (C) 2004  Hongli Lai <h.lai@chello.nl>
 *                      Mike Hearn <mike@navi.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#ifndef __USE_GNU
#	define __USE_GNU
#endif
#ifndef __USE_XOPEN
#	define __USE_XOPEN
#endif
#include <stdlib.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <termios.h>
#include <string.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include "autosu.h"

static gboolean exists(char *path) {
        struct stat buf;
        if (stat(path, &buf) == -1) return FALSE;
        return TRUE;
}

static char *get_xauth_path() {
#define check(p) if (exists(p)) return p;
        
        check("/usr/X11R6/bin/xauth");
        check("/usr/X11/bin/xauth");

        // fixme: what else goes here?
        perror("could not locate xauth");
        exit(1);
#undef check
}

/* return a shmid for a segment containing the X authentication tokens for this system */
static int load_xauth()
{
        char *cmd = g_strdup_printf("%s list", get_xauth_path());
        FILE *in = popen(cmd, "r");
        free(cmd);
        if (!in) {
                perror("could not run xauth");
                exit(1);
        }
        
        static const int bufsize = 1024 * 5;
        void *buffer = malloc(bufsize); /* unlikely a token line will go >5k chars */
        GString *str = g_string_new("");
        
        while (!feof(in) && fgets(buffer, bufsize, in)) {
                g_string_append(str, "add ");
                g_string_append(str, buffer);
        }
        
        if (pclose(in) != 0) {
                perror("could not run xauth, pclose != 0");
                exit(1);
        }

        /* now we need to move the buffer to a shared memory segment */
        int shmid = shmget(IPC_PRIVATE, str->len, IPC_CREAT | IPC_EXCL | 0666);
        if (shmid == -1) {
                perror("could not generate shared memory segment for X authentication tokens");
                exit(1);
        }
        
        char *deststr = shmat(shmid, NULL, 0);
        if (deststr == (char*)-1) {
                perror("could not attach to shared memory segment for X authentication tokens");
                exit(1);
        }

        strcpy(deststr, str->str);
        g_string_free(str, TRUE);
        return shmid;
}

/* Functions for communicating with su */
static char *su_bin = NULL;

/* Create a pseudo terminal */
static char *
init_tty (int *amaster, int *aslave)
{
	int master, slave;
	char *tty;

	master = getpt ();
	if (master < 0)
	{
		show_error (_("Unable to create master pseudo terminal: %s."),
			g_strerror (errno));
		exit (EXIT_SYSERROR);
	}

	if (grantpt (master) != 0 || unlockpt (master) != 0 || !(tty = ptsname (master)))
	{
		close (master);
		show_error (_("Unable to allocate master pseudo terminal: %s."),
			g_strerror (errno));
		exit (EXIT_SYSERROR);
	}

	slave = open (tty, O_RDWR);
	if (!slave)
	{
		show_error (_("Unable to allocate slave pseudo terminal: %s."),
			g_strerror (errno));
		exit (EXIT_SYSERROR);
	}

	*amaster = master;
	*aslave = slave;
	return tty;
}


static void
attach_tty (int master, int slave)
{
	struct termios tio;

	close(master);
	dup2 (slave, 0);
	dup2 (slave, 1);
	dup2 (slave, 2);
	if (slave > 2)
		close (slave);

	tcgetattr (0, &tio);
	tio.c_oflag &= ~OPOST;
	tcsetattr (0, TCSANOW, &tio);
}


static void
su_init ()
{
	if (su_bin) return;
	if (g_file_test ("/bin/su", G_FILE_TEST_EXISTS))
		su_bin = "/bin/su";
	else if (g_file_test ("/usr/bin/su", G_FILE_TEST_EXISTS))
		su_bin = "/usr/bin/su";
	else
	{
		show_error (_("The required system program 'su' is not found."));
		exit (EXIT_SYSERROR);
	}
}

static gboolean
str_has_prefix (const gchar  *str,
                const gchar  *prefix)
{
  int str_len;
  int prefix_len;
  
  g_return_val_if_fail (str != NULL, FALSE);
  g_return_val_if_fail (prefix != NULL, FALSE);

  str_len = strlen (str);
  prefix_len = strlen (prefix);

  if (str_len < prefix_len)
    return FALSE;
  
  return strncmp (str, prefix, prefix_len) == 0;
}

static gboolean
str_has_suffix (const gchar  *str,
                const gchar  *suffix)
{
  int str_len;
  int suffix_len;
  
  g_return_val_if_fail (str != NULL, FALSE);
  g_return_val_if_fail (suffix != NULL, FALSE);

  str_len = strlen (str);
  suffix_len = strlen (suffix);

  if (str_len < suffix_len)
    return FALSE;

  return strcmp (str + str_len - suffix_len, suffix) == 0;
}


/* Check whether PAM is available. If not, exec(command) */
void
su_check_pam (char **argv)
{
	struct stat buf;

	su_init ();
	if (stat ("/etc/pam.d", &buf) != 0)
	{
		fprintf (stderr, "\e[1;31m");
		fprintf (stderr, "WARNING: ");
		fprintf (stderr, "\e[0m");
		fprintf (stderr, "your system doesn't seem to have PAM installed. autopackage normally displays a graphical password prompt, but that doesn't work if PAM is not installed. If you're a Slackware user, please install Dropline GNOME with PAM.\n\n");
		execvp (argv[0], argv);
	}
}


/* Check whether su requires a password for user. */
gboolean
su_needs_password (char *user)
{
	int master, slave;
	char *tty;
	pid_t pid;

	su_init ();
	tty = init_tty (&master, &slave);

	/* Fork and run su; attach pseudo terminal slave to it */
	pid = fork ();
	if (pid == -1)
	{
		show_error (_("An error has occured while trying to gain "
			"superuser privileges. Your system may be low on memory."
			"Please close a few applications to free some memory.\n\n"
			"Technical details: fork() failed: %s"),
			g_strerror (errno));
		exit (EXIT_SYSERROR);

	} else if (pid == 0)
	{
		/* Child */
		attach_tty (master, slave);
		setenv ("LC_ALL", "C", 1);
		execl (su_bin, su_bin, user, "-c", "true", NULL);
		_exit (1);
		return FALSE;

	} else
	{
		/* Parent */
		int status;
		char buf[1024];
		gboolean result = FALSE;

		close (slave);
		memset (buf, 0, sizeof (buf));
		read (master, buf, sizeof (buf));

		if (str_has_prefix (buf, "Password:"))
		{
			result = TRUE;
		} else if (strcmp (buf, "This account is currently not available.\n") == 0)
		{
			show_error (_("The account '%s' is currently not available."),
				user);
			kill (pid, SIGTERM);
			waitpid (pid, &status, 0);
			exit (EXIT_SYSERROR);
		} else if (str_has_suffix (buf, " does not exist\n"))
		{
			show_error (_("The user '%s' does not exist."), user);
			kill (pid, SIGTERM);
			waitpid (pid, &status, 0);
			exit (EXIT_SYSERROR);
		}

		close (master);
		kill (pid, SIGTERM);
		waitpid (pid, &status, 0);

		return result;
	}
}


gboolean
su_check_password (char *user, char *password)
{
	int master, slave;
	char *tty;
	pid_t pid;

	su_init ();
	tty = init_tty (&master, &slave);

	/* Fork and run su; attach pseudo terminal slave to it */
	pid = fork ();
	if (pid == -1)
	{
		safefree (password);
		show_error (_("Unable to fork a process."));
		exit (EXIT_SYSERROR);

	} else if (pid == 0)
	{
		/* Child */
		safefree (password);
		attach_tty (master, slave);
		setenv ("LC_ALL", "C", 1);
		execl (su_bin, su_bin, user, "-c", "true", NULL);
		_exit (1);
		return FALSE;

	} else
	{
		/* Parent */
		int status;
		char buf[1024];
		struct termios tio;

		memset (buf, 0, sizeof (buf));
		read (master, buf, sizeof (buf));

		if (!str_has_prefix (buf, "Password:"))
		{
			/* ??????? */
			safefree (password);
                        char *e = g_strdup_printf(_("Error 1: An unknown error has occured. (%s)"), buf);
			show_error (e);
			kill (pid, SIGTERM);
			waitpid (pid, &status, 0);
			exit (EXIT_SYSERROR);
		}

		/* Wait until the terminal is set to no echo mode */
		while (1)
		{
			/* The child already died? */
			pid_t retval = waitpid (pid, &status, WNOHANG);
			if (retval == -1 || retval > 0 || tcgetattr (slave, &tio) < 0)
			{
				close (slave);
				close (master);
				safefree (password);
				show_error (_("The superuser (root) privilege provider terminated unexpectedly."));
				exit (EXIT_SYSERROR);

			}

			if (tio.c_lflag & ECHO)
				usleep (50000);	/* Sleep 50ms */
			else
				break;
		}
		close (slave);

		/* Now write the password */
		write (master, password, strlen (password));
		write (master, "\n", 1);

		/* Read output to make it not lock up */
		read (master, buf, sizeof (buf));

		waitpid (pid, &status, 0);
		close (master);

		if (!WIFEXITED (status))
		{
			show_error (_("The superuser (root) privilege provider terminated unexpectedly."));
			exit (EXIT_SYSERROR);
		} else
			return WEXITSTATUS(status) == 0;
	}
}


/* Run su with command and password. Return su's exit code. */
gint
su_exec (char *user, char *command, char *password)
{
	int master, slave;
	char *tty, *helper_exe[2], tmp[PATH_MAX + 1], *tmp2;
	pid_t pid;

	su_init ();
	memset (tmp, 0, sizeof (tmp));
	if (!realpath ("/proc/self/exe", tmp))
	{
		show_error (_("Unable to locate read from /proc/self/exe."));
		exit (EXIT_SYSERROR);
	}
	tmp2 = g_dirname (tmp);
	helper_exe[0] = g_strdup_printf ("%s/autosu-helper", tmp2);
	helper_exe[1] = NULL;
	g_free (tmp2);
	if (!g_file_test (helper_exe[0], G_FILE_TEST_EXISTS))
	{
		show_error (_("Unable to locate autosu-helper. Your autopackage installation may be damaged."));
		exit (EXIT_SYSERROR);
	}


	tty = init_tty (&master, &slave);

	/* Fork and run su; attach pseudo terminal slave to it */
	pid = fork ();
	if (pid == -1)
	{
		safefree (password);
		show_error (_("Unable to fork a process."));
		exit (EXIT_SYSERROR);

	} else if (pid == 0)
	{
		/* Child */
		char *in_tty[2], *out_tty[2], *err_tty[2], *su_command;

		in_tty[0] = ttyname (0);
		if (!in_tty[0]) in_tty[0] = "";
		out_tty[0] = ttyname (1);
		if (!out_tty[0]) out_tty[0] = "";
		err_tty[0] = ttyname (2);
		if (!err_tty[0]) err_tty[0] = "";

		in_tty[1] = NULL;
		out_tty[1] = NULL;
		err_tty[1] = NULL;

		safefree (password);
		attach_tty (master, slave);
		setenv ("LC_ALL", "C", 1);

                char *xauth = NULL;
                if (system("grep -q pam_xauth /etc/pam.d/su") != 0)
                        xauth = g_strdup_printf("--xauth-shm-id %d", load_xauth());
                
		su_command = g_strdup_printf ("%s %s %s %s %s %s",
                                              create_command (helper_exe),
                                              xauth ? xauth : "",
                                              create_command (in_tty),
                                              create_command (out_tty),
                                              create_command (err_tty),
                                              command);
                if (xauth) free(xauth);
                
		execl (su_bin, su_bin, "-c", su_command, NULL);
		_exit (1);
		return 1;

	} else
	{
		/* Parent */
		int status;
		char buf[1024];
		struct termios tio;

		/* No password is given; just wait until su has exited */
		if (!password)
		{
			close (slave);
			waitpid (pid, &status, 0);
			close (master);

			if (!WIFEXITED (status))
			{
				if (WIFSIGNALED (status))
					return 128 + WTERMSIG (status);
				else
					return EXIT_ABNORMAL;
			} else
				return WEXITSTATUS (status);
		}


		memset (buf, 0, sizeof (buf));
		read (master, buf, sizeof (buf));

		if (!str_has_prefix (buf, "Password:"))
		{
			/* ??????? */
			safefree (password);
                        char *e = g_strdup_printf(_("Error 1: An unknown error has occured. (%s)"), buf);
			show_error (e);
			kill (pid, SIGTERM);
			waitpid (pid, &status, 0);
			exit (EXIT_SYSERROR);
		}

		/* Wait until the terminal is set to no echo mode */
		while (1)
		{
			/* The child already died? */
			pid_t retval = waitpid (pid, &status, WNOHANG);
			if (retval == -1 || retval > 0 || tcgetattr (slave, &tio) < 0)
			{
				safefree (password);
				show_error (_("The superuser (root) privilege provider terminated unexpectedly."));
				exit (EXIT_SYSERROR);

			}

			if (tio.c_lflag & ECHO)
				usleep (50000);	/* Sleep 50ms */
			else
				break;
		}
		close (slave);

		/* Now write the password */
		write (master, password, strlen (password));
		write (master, "\n", 1);
		safefree (password);

		/* Read output to make it not lock up */
		read (master, buf, sizeof (buf));

		waitpid (pid, &status, 0);
		close (master);

		if (!WIFEXITED (status))
		{
			if (WIFSIGNALED (status))
				return 128 + WTERMSIG (status);
			else
				return EXIT_ABNORMAL;
		} else
			return WEXITSTATUS (status);
	}
}


/*************************
 * Utility functions
 *************************/

void
safefree (gchar *password)
{
	if (!password) return;
	memset (password, 0, strlen (password));
	g_free (password);
}


static void
replace_all (gchar **str, gchar *from, gchar *to)
{
	GString *newstr;
	gchar *found;

	g_return_if_fail (str != NULL);
	g_return_if_fail (from != NULL);
	g_return_if_fail (to != NULL);

	newstr = g_string_new (*str);
	found = strstr (newstr->str, from);
	while (found != NULL)
	{
		gint pos;

		pos = GPOINTER_TO_INT (found) - GPOINTER_TO_INT (newstr->str);
		g_string_erase (newstr, pos, strlen (from));
		g_string_insert (newstr, pos, to);
		found = GINT_TO_POINTER (GPOINTER_TO_INT (found) + strlen (to));
		found = strstr (found, from);
	}

	g_free (*str);
	*str = newstr->str;
	g_string_free (newstr, FALSE);
}


/* Changes an argument vector into a commandline.
   Escape special characters. */
gchar *
create_command (gchar **argv)
{
	GString *result;
	gchar *str;
	guint i;

	result = g_string_new ("");
	i = 0;
	while (argv[i] != NULL)
	{
		gchar *tmp;

		tmp = g_strdup (argv[i]);
		replace_all (&tmp, "\"", "\\\"");
		replace_all (&tmp, ">", "\\>");
		replace_all (&tmp, "<", "\\<");
		replace_all (&tmp, " ", "\\ ");
		replace_all (&tmp, "!", "\\!");
		replace_all (&tmp, "$", "\\$");
		replace_all (&tmp, "&", "\\&");
		if (!*tmp)
			g_string_append (result, "\"\" ");
		else
			g_string_append_printf (result, "%s ", tmp);
		g_free (tmp);
		i++;
	}

	str = result->str;
	str[strlen (str) - 1] = 0; /* Get rid of trailing whitespace */
	g_string_free (result, FALSE);
	return str;
}


static void
usage (char *exe)
{
	fprintf (stderr, _("Usage: %s [--root-only] [-m apkg-install] [-d DISPLAYNAME] COMMAND [ARGS...]\n"), exe);
	exit (EXIT_SYSERROR);
}


char **
parse_args (int argc, char **argv, Options *options)
{
	char **command;
	int i;

        options->displayName = getenv("AUTOSU_DISPLAYNAME");
        
	if (argc <= 1)
		usage (argv[0]);

	for (i = 1; i < argc; i++)
	{
		if (strcmp (argv[i], "-m") == 0)
		{
			if (!argv[i + 1])
				usage (argv[0]);
			if (strcmp (argv[i + 1], "apkg-install") == 0)
				options->mode = APKG_INSTALL;
			else if (strcmp (argv[i + 1], "install") == 0)
				options->mode = INSTALL;
			else
				usage (argv[0]);
			i += 2;
		} else if (strcmp (argv[i], "--root-only") == 0)
                {
                        if (!argv[i + 1])
                                usage (argv[0]);
                        options->rootonly = TRUE;
                        i++;
                }
		break;
	}
	command = argv + i;

	return command;
}
