/*
 * luau (Lib Update/Auto-Update): Simple Update Library
 * Copyright (C) 2003  David Eklund
 *
 * - This library is free software; you can redistribute it and/or             -
 * - modify it under the terms of the GNU Lesser General Public                -
 * - License as published by the Free Software Foundation; either              -
 * - version 2.1 of the License, or (at your option) any later version.        -
 * -                                                                           -
 * - This library is distributed in the hope that it will be useful,           -
 * - but WITHOUT ANY WARRANTY; without even the implied warranty of            -
 * - MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         -
 * - Lesser General Public License for more details.                           -
 * -                                                                           -
 * - You should have received a copy of the GNU Lesser General Public          -
 * - License along with this library; if not, write to the Free Software       -
 * - Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA -
 */

/* TODO: use the gerrors we define */

#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include <errno.h>

#include <glib.h>
#include <curl/curl.h>
#include <curl/types.h>
#include <curl/easy.h>

#include "libuau.h"
#include "util.h"

#ifdef WITH_DMALLOC
#  include <dmalloc.h>
#endif

void ERROR(char *s, ...) {
	va_list args;
	va_start(args, s);
	fprintf(stderr, "ERROR: ");
	vfprintf(stderr, s, args);
	fprintf(stderr, "\n");
}

static int downloadURLUpdate(char *url, const char *updateName, const AInterface *interface, const char *version, const char *output, gboolean lookForMeta);
static char* downloadUpdate(AProgInfo *progInfo, AUpdate *updateInfo, APkgType pkgtype, const char *output, gboolean lookForMeta);
static char* downloadFile(const char *url, const char *output);

static AUpdate* getRightUpdate(const AProgInfo *progInfo, const AInterface *interface, const char *version, const char *updateName);
static GList* getRelevantUpdates(const AProgInfo *progInfo, APkgType type);

static gboolean curlFetch(const char *url, const char *loc);
static gboolean curlExists(const char *url);

static int progress_callback(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow);
static size_t autofail_callback(void *ptr, size_t size, size_t nmemb, void *stream);

static void printUsage(void);
static struct option* getLongOptions();

static char *pipeFile = "/dev/stdin";

void releaseBlock () {
	char inbuf[25];
	FILE *f = fopen(pipeFile, "r");
	if (fgets(inbuf, sizeof(inbuf), f) == NULL)
		ERROR("read failed");
	fclose(f);
}

int
main(int argc, char *argv[]) {
	struct option *longOptions = getLongOptions();
	gboolean lookForMeta = FALSE;
	char *program = NULL, *updates = NULL, *name = NULL, *file = NULL, *output = NULL, *version = NULL;
	AInterface *interface = NULL;
	int c, ret = 0;
	
	if (argc < 2) {
		printUsage();
		g_free(longOptions);
		exit(0);
	}
	
	luau_registerProgressCallback(progress_callback);
	
	while ((c = getopt_long(argc, argv, ":P:p:u:n:f:o:i:v:mh", longOptions, NULL)) != -1) {
		switch (c) {
			case 'P':
				pipeFile = g_strdup(optarg);
				break;
			case 'u':
				updates = g_strdup(optarg);
				break;
			case 'n':
				name = g_strdup(optarg);
				break;
			case 'f':
				file = g_strdup(optarg);
				break;
			case 'o':
				output = g_strdup(optarg);
				break;
			case 'i':
				interface = (AInterface*) g_malloc(sizeof(AInterface));
				sscanf(optarg, "%d.%d", &(interface->major), &(interface->minor));
				break;
			case 'v':
				version = g_strdup(optarg);
				break;
			case 'm':
				lookForMeta = TRUE;
				break;
			case 'h':
				printUsage();
				g_free(longOptions);
				exit(0);
			case '?':
				printf("Invalid option: -%c\n", optopt);
				printUsage();
				g_free(longOptions);
				exit(1);
			case ':':
				printf("Argument needed for -%c\n", optopt);
				printUsage();
				g_free(longOptions);
				exit(1);
			default:
				printf("Unknown error: got character code 0%o\n", c);
				printUsage();
				g_free(longOptions);
				exit(1);
		}
	}
	
	g_free(longOptions);

	ret = 1;
	if (argc == optind) {
		if (updates != NULL) {
			if (file != NULL)
				ERROR("Only one of --file, --updates, or --program can be specified");
			else
				ret = downloadURLUpdate(updates, name, interface, version, output, lookForMeta);
		}
		
		else if (file != NULL) {
			char *result;
			result = downloadFile(file, output);
			ret = (result != NULL);
			g_free(result);
		}
	} else {
		ERROR("Too many arguments");
		printUsage();
		ret = 1;
	}
	
	if (program != NULL) g_free(program);
	if (updates != NULL) g_free(updates);
	if (name    != NULL) g_free(name);
	if (file    != NULL) g_free(file);
	if (output  != NULL) g_free(output);
	if (interface != NULL) g_free(interface);

	return ret;
}


static int
downloadURLUpdate(char *url, const char *updateName, const AInterface *interface, const char *version, const char *output, gboolean lookForMeta) {
	AProgInfo progInfo;
	AUpdate *updateInfo;
	char *loc;

	memset(&progInfo, 0, sizeof(AProgInfo));
	progInfo.url = url;
	updateInfo = getRightUpdate(&progInfo, interface, version, updateName);
	
	if (updateInfo == NULL) {
		ERROR("No download was available that matched the requested constraints");
		return 1;
	}
	
	loc = downloadUpdate(&progInfo, updateInfo, LUAU_AUTOPKG, output, lookForMeta);
	
	luau_freeUpdateInfo(updateInfo);
	g_free(updateInfo);
	
	return (loc == NULL);
}

static char *
downloadUpdate(AProgInfo *progInfo, AUpdate *updateInfo, APkgType pkgtype, const char *output, gboolean lookForMeta) {
	APackage *pkgInfo;
	char *loc, *url, *url2;
        GError *error = NULL;

	pkgInfo = luau_getUpdatePackage(updateInfo, pkgtype, &error);
	if (pkgInfo == NULL) {
		ERROR("Specified format not available for selected update");
		return NULL;
	}
	url = luau_getPackageURL(pkgInfo, &error);
	url2 = g_strdup_printf("%s.%s", url, lookForMeta ? "meta" : "payload");
	  
	loc = downloadFile(url2, output);
	
	//luau_freePkgInfo(pkgInfo);  - double frees here
	g_free(url2);
	//g_free(url);
	
	return loc;
}

/* We need to open/write/close each time because this is how bash does it when using echo/cat etc,
   and the frontend/backend sync can break if we don't do that. */
void out(char *s, ...) {
	va_list args;
	FILE *f = fopen(pipeFile, "w");
	if (!f) {
		ERROR("out: could not open %s", pipeFile);
		exit(1);
	}
	va_start(args, s);
	vfprintf(f, s, args);
	va_end(args);

	fclose(f);
}

static char *
downloadFile(const char *url, const char *output) {
	char *loc, *realURL;
	int len, i;
	const char *filename;
	gboolean result;
	
	realURL = NULL;
	
	if (output == NULL || output[0] == '\0')
		output = "/tmp";
	
	if (!curlExists(url)) {
		ERROR("url %s was requested but not found: aborting", url);
		return NULL;
	}
	
	if (lutil_isDirectory(output)) {
		filename = strrchr(url, '/') + 1;
		if (filename == NULL || filename[0] == '\0') {
			ERROR("Invalid URL given to download (%s): aborting", url);
			return NULL;
		}
		loc = lutil_vstrcreate(output, "/", filename, NULL);
		
		len = strlen(output) + strlen(filename) + 8;
		for (i = 1; lutil_fileExists(loc); ++i) {
			g_free(loc);
			loc = lutil_createString(len);
			snprintf(loc, len+1, "%s/%s.%d", output, filename, i);
		}
	}
	else
		loc = g_strdup(output);


	/* we need to process url to get just the filename, to keep it at a manageable length */
	char *urldesc = strdup(url);
	if (strrchr(urldesc, '/')) urldesc = strrchr(urldesc, '/');
	out("DOWNLOAD-START\n%s\nDescription?\n", ++urldesc);
	
	releaseBlock();
	
	out("DOWNLOAD-PREPARE\nCONNECTING\n\n");

	releaseBlock();
	
	result = curlFetch(url, loc);
	if (result == FALSE) {
		g_free(loc);
		loc = NULL;
		
		out("DOWNLOAD-FINISH\nFAILURE\n\n");
		releaseBlock();
	} else {
		out("DOWNLOAD-FINISH\nSUCCESS\n\n");
		releaseBlock();
	}
	
	if (realURL != NULL)
		g_free(realURL);
	
	return loc;
}

static AUpdate *
getRightUpdate(const AProgInfo *progInfo, const AInterface *interface, const char *version, const char *updateName) {
        AUpdate *update = NULL;
        GList *allUpdates, *current;
        GError *err;
        
        if (updateName != NULL) {
                update = g_malloc(sizeof(AUpdate));
                luau_getUpdateInfo(update, updateName, progInfo, &err);
                return update;
        } else if (version != NULL) {
                allUpdates = getRelevantUpdates(progInfo, LUAU_AUTOPKG);
                if (!allUpdates) return NULL;

                current = allUpdates;

                do {
                        update = current->data;
                        g_assert( update != NULL );

                        /* if we found a good version, return it */
                        if (update->newVersion && lutil_streq(update->newVersion, version)) {
                                allUpdates = g_list_remove_link(allUpdates, current);
                                luau_freeUpdateList(allUpdates);
                                return update;
                        }
                } while ((current = g_list_next(current)));

                return NULL;
        } else if (interface != NULL) {
                AUpdate *candidate = NULL;
                
                allUpdates = getRelevantUpdates(progInfo, LUAU_AUTOPKG);
                if (!allUpdates) return NULL;

                current = allUpdates;
                do {
                        update = current->data;
                        if (!luau_satisfiesInterface(&update->interface, interface)) continue;
                        if (!candidate || (update->interface.minor > candidate->interface.minor))
                                candidate = update;
                } while ((current = g_list_next(current)));

                return candidate;
        } else {
                AUpdate *candidate = NULL;
                
                allUpdates = getRelevantUpdates(progInfo, LUAU_AUTOPKG);
                if (!allUpdates) return NULL;
                
                current = allUpdates;
                do {
                        update = current->data;
                        if (!candidate || (update->date && luau_datecmp(update->date, candidate->date)))
                                candidate = update;
                } while ((current = g_list_next(current)));

                g_assert( candidate != NULL );
                
                allUpdates = g_list_remove(allUpdates, candidate);
                luau_freeUpdateList(allUpdates);
                        
                return candidate;
        }
}

static GList*
getRelevantUpdates(const AProgInfo *progInfo, APkgType type) {
	GList *updates = NULL;
        GList *current;
	AUpdate *update;
	GError *error = NULL;
	
	updates = luau_checkForUpdates(progInfo, &error);
	if (updates == NULL) {
		ERROR("Couldn't retrieve updates for program");
		return NULL;
	}

        current = updates;

        /* eliminate unsuitable updates from the list */
        do {
                
                update = (AUpdate *)current->data;
                
                if ((update->type != LUAU_SOFTWARE) || !luau_isOfType(update->availableFormats, type)) {
                        luau_freeUpdateInfo(update);
			g_free(update);
                        current->data = NULL;
		}
                
        } while ((current = g_list_next(current)));

	return g_list_remove_all(updates, NULL);
}
	
static gboolean
curlFetch(const char *url, const char *loc) {
	CURL *curl_handle;
	CURLcode ret;
	FILE *tempFile;
	time_t currentTime = time(NULL);
	char errorBuffer[CURL_ERROR_SIZE];
	
	tempFile = fopen(loc, "w");
	if (tempFile == NULL) {
		ERROR("Couldn't download file to specified location %s: %s", loc, strerror(errno));
		return FALSE;
	}
	
	curl_handle = curl_easy_init();

	curl_easy_setopt(curl_handle, CURLOPT_URL, url);
	curl_easy_setopt(curl_handle, CURLOPT_FILE, (void *)tempFile);
	curl_easy_setopt(curl_handle, CURLOPT_ERRORBUFFER, errorBuffer);
	curl_easy_setopt(curl_handle, CURLOPT_CONNECTTIMEOUT, 15);
	
	curl_easy_setopt(curl_handle, CURLOPT_NOPROGRESS, FALSE);
	curl_easy_setopt(curl_handle, CURLOPT_PROGRESSFUNCTION, progress_callback);
	curl_easy_setopt(curl_handle, CURLOPT_PROGRESSDATA, &currentTime);
	
	curl_easy_setopt(curl_handle, CURLOPT_FAILONERROR, 1);
		
	ret = curl_easy_perform(curl_handle);

	curl_easy_cleanup(curl_handle);
	
	fclose(tempFile);
	
	if (ret != 0) {
		unlink(loc);
		ERROR("Couldn't download %s: %s", url, errorBuffer);
		return FALSE;
	} else {
		return TRUE;
	}
}


static gboolean
curlExists(const char *url) {
	CURL *curl_handle;
	CURLcode ret;
	
	curl_handle = curl_easy_init();

	curl_easy_setopt(curl_handle, CURLOPT_URL, url);
	curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, autofail_callback);
	curl_easy_setopt(curl_handle, CURLOPT_CONNECTTIMEOUT, 15);
	curl_easy_setopt(curl_handle, CURLOPT_FAILONERROR, 1);
	
	ret = curl_easy_perform(curl_handle);

	curl_easy_cleanup(curl_handle);
	
	/* If the connection is made and the file exists, the write function 'autofail_callback' will be
	   called (which then signals an error and causes libcurl to return a write error ("CURLE_WRITE_ERROR").
	   Otherwise, some other error will be returned, meaning the file cannot be retrieved. */
	return (ret == CURLE_WRITE_ERROR);
}


static int
progress_callback(void *clientp, double dltotal, double dlnow, double ultotal, double ulnow) {
	int timeRem, timePast = 0;
	float rate;
	static int last_dl = 0;

	GString *buf = g_string_new("DOWNLOAD-PROGRESS\n");
	rate = (dlnow - last_dl) / 1024;
	last_dl = dlnow;
/*  	timePast = time(NULL) - *start_time; */
/* 	if (timePast > 0) */
/* 		rate = (dlnow / 1024) / timePast; */
/* 	else */
/* 		rate = 0; */
	g_string_append_printf(buf, "%d\n", (int) rate);
	
	if ((dltotal != 0) && (dltotal >= dlnow) && (rate > 0)) {
		g_string_append_printf(buf, "%d\n", (int) ((dlnow / dltotal) * 100) );
		timeRem = (dltotal - dlnow) / (1024 * rate);
		g_string_append_printf(buf, "%02d:%02d\n", (int) (timeRem / 60), (timeRem % 60));
	} else if (dlnow == 0 || timePast == 0) {
		g_string_append_printf(buf, "0\n");
 		g_string_append_printf(buf, "??:??\n");
	} else {
		g_string_append_printf(buf, "?\n");
		g_string_append_printf(buf, "??:??\n");
	}
	out(buf->str);
	g_string_free(buf, TRUE);
	releaseBlock();
	return 0;
}

static size_t autofail_callback(void *ptr, size_t size, size_t nmemb, void *stream) {
	return 0; /* causes libcurl to disconnect with CURLE_WRITE_ERROR */
}


static void
printUsage(void) {
	printf("Usage: luau-download [OPTION] ...\n");
	printf("\n");
	printf("Download Luau provided files.\n");
	printf("\n");
	printf("Options:\n");
	printf("  -u, --updates='URL'       updates URL to use\n");
	printf("  -f, --file='URL'          download the specified file\n\n");
	
	printf("  -n, --name='NAME'         name of update to download\n");
	printf("  -i, --interface='VERSION' specify an interface version to conform to\n");
	printf("  -v, --version='VERSION'   specify an exact version to download\n\n");
	
	printf("  -m, --meta                check for and download .meta file, if available\n\n");
	
	printf("  -o, --output='PATH'       file or directory to download to\n");
	printf("  -h, --help                display this message\n");
	printf("\n");
	printf("  The options --program, --url, and --file are mutually exclusive.  Use the\n");
	printf("  first to download files for an already installed program; use the second to\n");
	printf("  download files where you know the location of the updates file; and use the\n");
	printf("  last when you know the exact URL of the file you want.  The --name or\n");
	printf("  --interface option may be used with --program or --url, but never with --file;\n");
	printf("  likewise, the --meta option is optional for --program or --url, but may not\n");
	printf("  be used with --file.\n");
	printf("\n");
}

static struct option *
getLongOptions() {
	struct option *options = (struct option *) calloc(11, sizeof(struct option));
	
	options[0].name = "version";
	options[0].has_arg = 1;
	options[0].flag = NULL;
	options[0].val = 'v';
		
	options[1].name = "updates";
	options[1].has_arg = 1;
	options[1].flag = NULL;
	options[1].val = 'u';
	
	options[2].name = "name";
	options[2].has_arg = 1;
	options[2].flag = NULL;
	options[2].val = 'n';
	
	options[3].name = "m";
	options[3].has_arg = 0;
	options[3].flag = NULL;
	options[3].val = 'm';
	
	options[4].name = "file";
	options[4].has_arg = 1;
	options[4].flag = NULL;
	options[4].val = 'f';
	
	options[5].name = "help";
	options[5].has_arg = 0;
	options[5].flag = NULL;
	options[5].val = 'h';
	
	options[6].name = "output";
	options[6].has_arg = 1;
	options[6].flag = NULL;
	options[6].val = 'o';
	
	options[7].name = "interface";
	options[7].has_arg = 1;
	options[7].flag = NULL;
	options[7].val = 'i';
	

	memset(&options[8], 0, sizeof(struct option));
	
	return options;
}

